#
# Copyright (C) 2021  Nexedi SA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from collections import defaultdict, deque
from contextlib import contextmanager, nested
from time import time
from persistent import Persistent
from ZODB.POSException import UndoError
from neo.lib.protocol import Packets
from neo.lib.util import timeFromTID
from .. import consume, Patch
from . import ConnectionFilter, NEOThreadedTest, with_cluster

class PCounter(Persistent):
    value = 0

class PackTests(NEOThreadedTest):

    @contextmanager
    def assertPackOperationCount(self, cluster_or_storages, *counts):
        def patch(storage):
            packs = {}
            def _pack(orig, *args):
                offset = args[0]
                tid = args[2]
                try:
                    tids = packs[offset]
                except KeyError:
                    packs[offset] = [tid]
                else:
                    self.assertLessEqual(tids[-1], tid)
                    tids.append(tid)
                return orig(*args)
            return Patch(storage.dm, _pack=_pack), packs
        patches, packs = zip(*map(patch,
            getattr(cluster_or_storages, 'storage_list', cluster_or_storages)))
        with nested(*patches):
            yield
        self.assertSequenceEqual(counts,
            tuple(sum(len(set(x)) for x in x.itervalues()) for x in packs))

    def countAskPackOrders(self, connection_filter):
        counts = defaultdict(int)
        @connection_filter.add
        def _(conn, packet):
            if isinstance(packet, Packets.AskPackOrders):
                counts[self.getConnectionApp(conn).uuid] += 1
        return counts

    def populate(self, cluster):
        t, c = cluster.getTransaction()
        r = c.root()
        for x in 'ab', 'ac', 'ab', 'bd', 'c':
            for x in x:
                try:
                    r[x].value += 1
                except KeyError:
                    r[x] = PCounter()
            t.commit()
            yield cluster.client.last_tid
        c.close()

    @with_cluster(partitions=3, replicas=1, storage_count=3)
    def testOutdatedNodeIsBack(self, cluster):
        s0 = cluster.storage_list[0]
        populate = self.populate(cluster)
        consume(populate, 3)
        with self.assertPackOperationCount(cluster, 0, 4, 4), \
             ConnectionFilter() as f:
            counts = self.countAskPackOrders(f)
            with s0.delayTasks(1):
                cluster.client.pack(time())
                consume(populate, 2)
                cluster.client.pack(time())
                s0.stop()
                cluster.join((s0,))
            self.tic()
        # First storage node stops any pack-related work after the first
        # response to AskPackOrders. Other storage nodes process a pack order
        # for all cells before asking the master for the next pack order.
        self.assertEqual(counts, {s.uuid: 1 if s is s0 else 2
                                  for s in cluster.storage_list})
        s0.resetNode()
        with self.assertPackOperationCount(cluster, 4, 0, 0), \
             ConnectionFilter() as f:
            counts = self.countAskPackOrders(f)
            deque(populate, 0)
            s0.start()
            self.tic()
        # The master queries 2 storage nodes for old pack orders and remember
        # those that s0 has not completed. s0 processes all orders for the first
        # replicated cell and ask them again when the second is up-to-date.
        self.assertEqual(counts, {s0.uuid: 4, cluster.master.uuid: 2})
        self.checkReplicas(cluster)

    @with_cluster(replicas=1)
    def testValueSerialVsReplication(self, cluster):
        t, c = cluster.getTransaction()
        ob = c.root()[''] = PCounter()
        t.commit()
        s0 = cluster.storage_list[0]
        s0.stop()
        cluster.join((s0,))
        ob.value += 1
        t.commit()
        ob.value += 1
        t.commit()
        s0.resetNode()
        with ConnectionFilter() as f:
            f.delayAskFetchTransactions()
            s0.start()
            c.db().undo(ob._p_serial, t.get())
            t.commit()
            cluster.client.pack(time())
            self.tic()
        self.tic()
        self.checkReplicas(cluster)

    @with_cluster()
    def _testValueSerialMultipleUndo(self, cluster, *undos):
        t, c = cluster.getTransaction()
        r = c.root()
        ob = r[''] = PCounter()
        t.commit()
        tids = []
        for x in xrange(2):
            ob.value += 1
            t.commit()
            tids.append(ob._p_serial)
        r._p_changed = 1
        t.commit()
        db = c.db()
        def undo(i):
            db.undo(tids[i], t.get())
            t.commit()
            tids.append(db.lastTransaction())
        undo(-1)
        for i in undos:
            undo(i)
        cluster.client.pack(timeFromTID(r._p_serial))
        self.tic()
        undo(2)

    def testValueSerialMultipleUndo1(self):
        self._testValueSerialMultipleUndo(0, -1)

    def testValueSerialMultipleUndo2(self):
        self._testValueSerialMultipleUndo(-1, 1)


if __name__ == "__main__":
    unittest.main()
